#include "sprites.h"
#include "serviceLocator.h"
#include "d2d.h"

#include <wincodec.h>	// Windows Imaging Component
#include <WTypes.h>

namespace graphics
{
	/////////////////////////////////////////////////////////////////////////////////////////
	////////////////////////////// Constructor //////////////////////////////////////////////
	/////////////////////////////////////////////////////////////////////////////////////////
	Sprite::Sprite(Direct2D* d2d, LPCWSTR imageFile, float x, float y, Layers layer, unsigned int drawOrder) : layer(layer),
																						 drawOrder(drawOrder), x(x), y(y)
	{
		this->d2d = d2d;

		// create decoder
		Microsoft::WRL::ComPtr<IWICBitmapDecoder> bitmapDecoder;
		if (FAILED(d2d->WICFactory->CreateDecoderFromFilename(imageFile, NULL, GENERIC_READ, WICDecodeMetadataCacheOnLoad, bitmapDecoder.ReleaseAndGetAddressOf())))
			throw std::runtime_error("Failed to create decoder from filename!");

		// get the correct frame
		Microsoft::WRL::ComPtr<IWICBitmapFrameDecode> frame;
		if (FAILED(bitmapDecoder->GetFrame(0, frame.ReleaseAndGetAddressOf())))
			throw std::runtime_error("Failed to retrieve frame from bitmap!");

		// create the format converter
		Microsoft::WRL::ComPtr<IWICFormatConverter> image;
		if (FAILED(d2d->WICFactory->CreateFormatConverter(image.ReleaseAndGetAddressOf())))
			throw std::runtime_error("Failed to create the format converter!");

		// initialize the WIC image
		if (FAILED(image->Initialize(frame.Get(), GUID_WICPixelFormat32bppPBGRA, WICBitmapDitherTypeNone, NULL, 0, WICBitmapPaletteTypeCustom)))
			throw std::runtime_error("Failed to initialize the WIC image!");

		// create the bitmap
		if (FAILED(d2d->devCon->CreateBitmapFromWicBitmap(image.Get(), this->bitmap.ReleaseAndGetAddressOf())))
			throw std::runtime_error("Failed to create the bitmap image!");
	}

	SpriteMap::SpriteMap()
	{

	}

	Sprite::~Sprite()
	{
		// release bitmap
		this->bitmap.ReleaseAndGetAddressOf();
	}

	SpriteMap::~SpriteMap()
	{
		for (auto sprite : backgroundMap)
			delete sprite.second;
		for (auto sprite : characterMap)
			delete sprite.second;
		for (auto sprite : userInterfaceMap)
			delete sprite.second;
	}

	/////////////////////////////////////////////////////////////////////////////////////////
	/////////////////////////////// Populating //////////////////////////////////////////////
	/////////////////////////////////////////////////////////////////////////////////////////
	void SpriteMap::addSprite(Sprite* sprite)
	{
		Layers layer = sprite->layer;
		switch (layer)
		{
		case Layers::Background:
			backgroundMap.insert(std::make_pair<unsigned int, Sprite*>(std::move(sprite->drawOrder), std::move(sprite)));
			break;

		case Layers::Characters:
			characterMap.insert(std::make_pair<unsigned int, Sprite*>(std::move(sprite->drawOrder), std::move(sprite)));
			break;

		case Layers::UserInterface:
			userInterfaceMap.insert(std::make_pair<unsigned int, Sprite*>(std::move(sprite->drawOrder), std::move(sprite)));
			break;
		}
	}

	util::Expected<void> SpriteMap::addSprite(Direct2D* d2d, LPCWSTR imageFile, float x, float y, Layers layer, unsigned int drawOrder)
	{
		try
		{
			Layers l = layer;

			switch (l)
			{
			case Layers::Background:
				backgroundMap.insert(std::make_pair<unsigned int, Sprite*>(std::move(drawOrder), new Sprite(d2d, imageFile, x, y, layer, drawOrder)));
				break;

			case Layers::Characters:
				characterMap.insert(std::make_pair<unsigned int, Sprite*>(std::move(drawOrder), new Sprite(d2d, imageFile, x, y, layer, drawOrder)));
				break;

			case Layers::UserInterface:
				userInterfaceMap.insert(std::make_pair<unsigned int, Sprite*>(std::move(drawOrder), new Sprite(d2d, imageFile, x, y, layer, drawOrder)));
				break;
			}
		}
		catch (std::runtime_error& e)
		{
			return e;
		}

		// return success
		return { };
	}

	/////////////////////////////////////////////////////////////////////////////////////////
	////////////////////////////////// Drawing //////////////////////////////////////////////
	/////////////////////////////////////////////////////////////////////////////////////////
	void Sprite::draw(D2D1_RECT_F* destRect, D2D1_RECT_F* sourceRect, float opacity, D2D1_BITMAP_INTERPOLATION_MODE interPol)
	{
		if (!destRect)
		{
			// get size
			D2D1_SIZE_U size = this->bitmap->GetPixelSize();

			// set suitable destination rectangle
			D2D1_RECT_F rect = { this->x, this->y, this->x + size.width, this->y + size.height };

			d2d->devCon->DrawBitmap(bitmap.Get(), rect, opacity, interPol, sourceRect);
		}
		else
			d2d->devCon->DrawBitmap(bitmap.Get(), destRect, opacity, interPol, sourceRect);
	}

	void SpriteMap::draw(D2D1_RECT_F* destRect, D2D1_RECT_F* sourceRect, DrawCommands drawCommand, float opacity, D2D1_BITMAP_INTERPOLATION_MODE interPol)
	{
		D2D1_RECT_F* rect;
		if (destRect)
			rect = destRect;
		else
			rect = NULL;

		switch (drawCommand)
		{
		case DrawCommands::All:
			// draw the background first
			for (auto sprite : backgroundMap)
				sprite.second->draw(rect, sourceRect, opacity, interPol);

			// draw the characters
			for (auto sprite : characterMap)
				sprite.second->draw(rect, sourceRect, opacity, interPol);

			// draw the user interface
			for (auto sprite : userInterfaceMap)
				sprite.second->draw(rect, sourceRect, opacity, interPol);

			break;

		case DrawCommands::onlyBackground:
			// draw the background first
			for (auto sprite : backgroundMap)
				sprite.second->draw(rect, sourceRect, opacity, interPol);
			break;

		case DrawCommands::onlyCharacters:
			// draw the background first
			for (auto sprite : characterMap)
				sprite.second->draw(rect, sourceRect, opacity, interPol);
			break;

		case DrawCommands::onlyUserInterface:
			// draw the background first
			for (auto sprite : userInterfaceMap)
				sprite.second->draw(rect, sourceRect, opacity, interPol);
			break;
		}
	}

}