#pragma once

/****************************************************************************************
* Author:	Gilles Bellot
* Date:		28/05/2018 - Lenningen - Luxembourg
*
* Desc:		Sprites!

* History:
****************************************************************************************/

// INCLUDES /////////////////////////////////////////////////////////////////////////////

// C++ includes
#include <string.h>
#include <map>

// Windows and COM
#include <wrl/client.h>

// DirectX includes
#include <d2d1_3.h>

// bell0bytes utilities
#include "expected.h"

// CLASSES //////////////////////////////////////////////////////////////////////////////
namespace graphics
{
	// forward declarations
	class Direct2D;

	// layers are used in the painter's algorithm
	enum Layers { Background, Characters, UserInterface };

	// specify what to draw
	enum DrawCommands { All, onlyBackground, onlyCharacters, onlyUserInterface };

	class Sprite
	{
	private:
		Direct2D* d2d;										// pointer to the Direct2D class
		Microsoft::WRL::ComPtr<ID2D1Bitmap1> bitmap;		// a bitmap of the actual image
		Layers layer;										// the layer the bitmap belongs to
		unsigned int drawOrder;								// the draw order of the bitmap; relative to the layer
		float x, y;											// position of the bitmap

	public:
		// constructors and destructors
		Sprite(Direct2D* d2d, LPCWSTR imageFile, float x = 0.0f, float y = 0.0f, Layers layer = Layers::Characters, unsigned int drawOrder = 0);	// loads an image from the disk and saves it as a sprite
		~Sprite();

		// drawing
		void draw(D2D1_RECT_F* destRect, D2D1_RECT_F* sourceRect, float opacity = 1.0f, D2D1_BITMAP_INTERPOLATION_MODE interPol = D2D1_BITMAP_INTERPOLATION_MODE_NEAREST_NEIGHBOR);		// draws the sprite at the given location, with the given opacity and interpolation mode
		
		// friends
		friend class SpriteMap;
	};

	class SpriteMap
	{
	private:
		std::multimap<unsigned int, Sprite*> backgroundMap;		// holds the background sprites
		std::multimap<unsigned int, Sprite*> characterMap;		// holds the character sprites
		std::multimap<unsigned int, Sprite*> userInterfaceMap;	// holds the user interface sprites

	public:
		// constructors
		SpriteMap();
		~SpriteMap();

		// populate the sprite map
		void addSprite(Sprite* sprite);							// adds an existing sprite to its correct map
		util::Expected<void> addSprite(Direct2D* d2d, LPCWSTR imageFile, float x = 0.0f, float y = 0.0f, Layers layer = Layers::Characters, unsigned int drawOrder = 0);	// create a new sprite and adds it to sprite map
		
		// draw the sprites
		void draw(D2D1_RECT_F* destRect, D2D1_RECT_F* sourceRect, DrawCommands drawCommand = DrawCommands::All, float opacity = 1.0f, D2D1_BITMAP_INTERPOLATION_MODE interPol = D2D1_BITMAP_INTERPOLATION_MODE_NEAREST_NEIGHBOR);	// draw sprites based on layers and draw orders
	};
}